# https://cocodataset.org/#format-data

from concurrent.futures import ProcessPoolExecutor, as_completed
from typing import Any, Dict, List, Tuple, Optional
from pathlib import Path

import numpy as np
from tqdm import tqdm
from .... import pydantic_compat

from ...annotation_types import ImageData, MaskData, Mask, ObjectAnnotation, Label, Polygon, Point, Rectangle
from ...annotation_types.collection import LabelCollection
from .categories import Categories, hash_category_name
from .annotation import COCOObjectAnnotation, RLE, get_annotation_lookup, rle_decoding
from .image import CocoImage, get_image, get_image_id


def mask_to_coco_object_annotation(
        annotation: ObjectAnnotation, annot_idx: int, image_id: int,
        category_id: int) -> Optional[COCOObjectAnnotation]:
    # This is going to fill any holes into the multipolygon
    # If you need to support holes use the panoptic data format
    shapely = annotation.value.shapely.simplify(1).buffer(0)
    if shapely.is_empty:
        return

    xmin, ymin, xmax, ymax = shapely.bounds
    # Iterate over polygon once or multiple polygon for each item
    area = shapely.area

    return COCOObjectAnnotation(
        id=annot_idx,
        image_id=image_id,
        category_id=category_id,
        segmentation=[
            np.array(s.exterior.coords).ravel().tolist()
            for s in ([shapely] if shapely.type == "Polygon" else shapely.geoms)
        ],
        area=area,
        bbox=[xmin, ymin, xmax - xmin, ymax - ymin],
        iscrowd=0)


def vector_to_coco_object_annotation(annotation: ObjectAnnotation,
                                     annot_idx: int, image_id: int,
                                     category_id: int) -> COCOObjectAnnotation:
    shapely = annotation.value.shapely
    xmin, ymin, xmax, ymax = shapely.bounds
    segmentation = []
    if isinstance(annotation.value, Polygon):
        for point in annotation.value.points:
            segmentation.extend([point.x, point.y])
    else:
        box = annotation.value
        segmentation.extend([
            box.start.x, box.start.y, box.end.x, box.start.y, box.end.x,
            box.end.y, box.start.x, box.end.y
        ])

    return COCOObjectAnnotation(id=annot_idx,
                                image_id=image_id,
                                category_id=category_id,
                                segmentation=[segmentation],
                                area=shapely.area,
                                bbox=[xmin, ymin, xmax - xmin, ymax - ymin],
                                iscrowd=0)


def rle_to_common(class_annotations: COCOObjectAnnotation,
                  class_name: str) -> ObjectAnnotation:
    mask = rle_decoding(class_annotations.segmentation.counts,
                        *class_annotations.segmentation.size[::-1])
    return ObjectAnnotation(name=class_name,
                            value=Mask(mask=MaskData.from_2D_arr(mask),
                                       color=[1, 1, 1]))


def segmentations_to_common(class_annotations: COCOObjectAnnotation,
                            class_name: str) -> List[ObjectAnnotation]:
    # Technically it is polygons. But the key in coco is called segmentations..
    annotations = []
    for points in class_annotations.segmentation:
        annotations.append(
            ObjectAnnotation(name=class_name,
                             value=Polygon(points=[
                                 Point(x=points[i], y=points[i + 1])
                                 for i in range(0, len(points), 2)
                             ])))
    return annotations


def object_annotation_to_coco(
        annotation: ObjectAnnotation, annot_idx: int, image_id: int,
        category_id: int) -> Optional[COCOObjectAnnotation]:
    if isinstance(annotation.value, Mask):
        return mask_to_coco_object_annotation(annotation, annot_idx, image_id,
                                              category_id)
    elif isinstance(annotation.value, (Polygon, Rectangle)):
        return vector_to_coco_object_annotation(annotation, annot_idx, image_id,
                                                category_id)
    else:
        return None


def process_label(
    label: Label,
    idx: int,
    image_root: str,
    max_annotations_per_image=10000
) -> Tuple[np.ndarray, List[COCOObjectAnnotation], Dict[str, str]]:
    annot_idx = idx * max_annotations_per_image
    image_id = get_image_id(label, idx)
    image = get_image(label, image_root, image_id)
    coco_annotations = []
    annotation_lookup = get_annotation_lookup(label.annotations)
    categories = {}
    for class_name in annotation_lookup:
        for annotation in annotation_lookup[class_name]:
            category_id = categories.get(annotation.name) or hash_category_name(
                annotation.name)
            coco_annotation = object_annotation_to_coco(annotation, annot_idx,
                                                        image_id, category_id)
            if coco_annotation is not None:
                coco_annotations.append(coco_annotation)
                if annotation.name not in categories:
                    categories[annotation.name] = category_id
                annot_idx += 1

    return image, coco_annotations, categories


class CocoInstanceDataset(pydantic_compat.BaseModel):
    info: Dict[str, Any] = {}
    images: List[CocoImage]
    annotations: List[COCOObjectAnnotation]
    categories: List[Categories]

    @classmethod
    def from_common(cls,
                    labels: LabelCollection,
                    image_root: Path,
                    max_workers=8):
        all_coco_annotations = []
        categories = {}
        images = []
        futures = []
        coco_categories = {}

        if max_workers:
            with ProcessPoolExecutor(max_workers=max_workers) as exc:
                futures = [
                    exc.submit(process_label, label, idx, image_root)
                    for idx, label in enumerate(labels)
                ]
                results = [
                    future.result() for future in tqdm(as_completed(futures))
                ]
        else:

            results = [
                process_label(label, idx, image_root)
                for idx, label in enumerate(labels)
            ]

        for result in results:
            images.append(result[0])
            all_coco_annotations.extend(result[1])
            coco_categories.update(result[2])

        category_mapping = {
            category_id: idx + 1
            for idx, category_id in enumerate(coco_categories.values())
        }
        categories = [
            Categories(id=category_mapping[idx],
                       name=name,
                       supercategory='all',
                       isthing=1) for name, idx in coco_categories.items()
        ]
        for annot in all_coco_annotations:
            annot.category_id = category_mapping[annot.category_id]

        return CocoInstanceDataset(info={'image_root': image_root},
                                   images=images,
                                   annotations=all_coco_annotations,
                                   categories=categories)

    def to_common(self, image_root):
        category_lookup = {
            category.id: category for category in self.categories
        }
        annotation_lookup = get_annotation_lookup(self.annotations)

        for image in self.images:
            im_path = Path(image_root, image.file_name)
            if not im_path.exists():
                raise ValueError(
                    f"Cannot find file {im_path}. Make sure `image_root` is set properly"
                )

            data = ImageData(file_path=str(im_path))
            annotations = []
            for class_annotations in annotation_lookup[image.id]:
                if isinstance(class_annotations.segmentation, RLE):
                    annotations.append(
                        rle_to_common(
                            class_annotations, category_lookup[
                                class_annotations.category_id].name))
                elif isinstance(class_annotations.segmentation, list):
                    annotations.extend(
                        segmentations_to_common(
                            class_annotations, category_lookup[
                                class_annotations.category_id].name))
            yield Label(data=data, annotations=annotations)



from copy import copy
import pytest
import labelbox.types as lb_types
from labelbox.data.serialization import NDJsonConverter
from labelbox.data.serialization.ndjson.objects import NDDicomSegments, NDDicomSegment, NDDicomLine
"""
Data gen prompt test data
"""

prompt_text_annotation = lb_types.PromptClassificationAnnotation(
                                    feature_schema_id="ckrb1sfkn099c0y910wbo0p1a",
                                    name="test",
                                    value=lb_types.PromptText(answer="the answer to the text questions right here"),
                                )

prompt_text_ndjson =  {
        "answer": "the answer to the text questions right here",
        "name": "test",
        "schemaId": "ckrb1sfkn099c0y910wbo0p1a",
        "dataRow": {
            "id": "ckrb1sf1i1g7i0ybcdc6oc8ct"
        },
    }

data_gen_label = lb_types.Label(
    data={"uid": "ckrb1sf1i1g7i0ybcdc6oc8ct"},
    annotations=[prompt_text_annotation]
)
     
"""
Prompt annotation test
"""

def test_serialize_label():
    serialized_label = next(NDJsonConverter().serialize([data_gen_label]))
    # Remove uuid field since this is a random value that can not be specified also meant for relationships
    del serialized_label["uuid"]
    assert serialized_label == prompt_text_ndjson


def test_deserialize_label():
    deserialized_label = next(NDJsonConverter().deserialize([prompt_text_ndjson]))
    if hasattr(deserialized_label.annotations[0], 'extra'):
        # Extra fields are added to deserialized label by default need removed to match
        deserialized_label.annotations[0].extra = {}
    assert deserialized_label.annotations == data_gen_label.annotations


def test_serialize_deserialize_label():
    serialized = list(NDJsonConverter.serialize([data_gen_label]))
    deserialized = next(NDJsonConverter.deserialize(serialized))
    if hasattr(deserialized.annotations[0], 'extra'):
        # Extra fields are added to deserialized label by default need removed to match
        deserialized.annotations[0].extra = {}
    print(data_gen_label.annotations)
    print(deserialized.annotations)
    assert deserialized.annotations == data_gen_label.annotations



from tempfile import NamedTemporaryFile
import uuid
from datetime import datetime
import json
import requests
import os

from unittest.mock import patch
import pytest

from labelbox.schema.media_type import MediaType
from labelbox import DataRow, AssetAttachment
from labelbox.exceptions import MalformedQueryException, ResourceCreationError, InvalidQueryError
from labelbox.schema.task import Task, DataUpsertTask
from labelbox.schema.data_row_metadata import DataRowMetadataField, DataRowMetadataKind

SPLIT_SCHEMA_ID = "cko8sbczn0002h2dkdaxb5kal"
TEST_SPLIT_ID = "cko8scbz70005h2dkastwhgqt"
TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"
CAPTURE_DT_SCHEMA_ID = "cko8sdzv70006h2dk8jg64zvb"
EXPECTED_METADATA_SCHEMA_IDS = [
    SPLIT_SCHEMA_ID, TEST_SPLIT_ID, TEXT_SCHEMA_ID, CAPTURE_DT_SCHEMA_ID
].sort()
CUSTOM_TEXT_SCHEMA_NAME = "custom_text"


@pytest.fixture
def mdo(client):
    mdo = client.get_data_row_metadata_ontology()
    try:
        mdo.create_schema(CUSTOM_TEXT_SCHEMA_NAME, DataRowMetadataKind.string)
    except MalformedQueryException:
        # Do nothing if already exists
        pass
    mdo._raw_ontology = mdo._get_ontology()
    mdo._build_ontology()
    yield mdo


@pytest.fixture
def conversational_content():
    return {
        'row_data': {
            "messages": [{
                "messageId": "message-0",
                "timestampUsec": 1530718491,
                "content": "I love iphone! i just bought new iphone! 🥰 📲",
                "user": {
                    "userId": "Bot 002",
                    "name": "Bot"
                },
                "align": "left",
                "canLabel": False
            }],
            "version": 1,
            "type": "application/vnd.labelbox.conversational"
        }
    }


@pytest.fixture
def tile_content():
    return {
        "row_data": {
            "tileLayerUrl":
                "https://s3-us-west-1.amazonaws.com/lb-tiler-layers/mexico_city/{z}/{x}/{y}.png",
            "bounds": [[19.405662413477728, -99.21052827588443],
                       [19.400498983095076, -99.20534818927473]],
            "minZoom":
                12,
            "maxZoom":
                20,
            "epsg":
                "EPSG4326",
            "alternativeLayers": [{
                "tileLayerUrl":
                    "https://api.mapbox.com/styles/v1/mapbox/satellite-streets-v11/tiles/{z}/{x}/{y}?access_token=pk.eyJ1IjoibWFwYm94IiwiYSI6ImNpejY4NXVycTA2emYycXBndHRqcmZ3N3gifQ.rJcFIG214AriISLbB6B5aw",
                "name":
                    "Satellite"
            }, {
                "tileLayerUrl":
                    "https://api.mapbox.com/styles/v1/mapbox/navigation-guidance-night-v4/tiles/{z}/{x}/{y}?access_token=pk.eyJ1IjoibWFwYm94IiwiYSI6ImNpejY4NXVycTA2emYycXBndHRqcmZ3N3gifQ.rJcFIG214AriISLbB6B5aw",
                "name":
                    "Guidance"
            }]
        }
    }


def make_metadata_fields():
    msg = "A message"
    time = datetime.utcnow()

    fields = [
        DataRowMetadataField(schema_id=SPLIT_SCHEMA_ID, value=TEST_SPLIT_ID),
        DataRowMetadataField(schema_id=CAPTURE_DT_SCHEMA_ID, value=time),
        DataRowMetadataField(schema_id=TEXT_SCHEMA_ID, value=msg),
    ]
    return fields


def make_metadata_fields_dict():
    msg = "A message"
    time = datetime.utcnow()

    fields = [{
        "schema_id": SPLIT_SCHEMA_ID,
        "value": TEST_SPLIT_ID
    }, {
        "schema_id": CAPTURE_DT_SCHEMA_ID,
        "value": time
    }, {
        "schema_id": TEXT_SCHEMA_ID,
        "value": msg
    }]
    return fields


def test_get_data_row_by_global_key(data_row_and_global_key, client, rand_gen):
    _, global_key = data_row_and_global_key
    data_row = client.get_data_row_by_global_key(global_key)
    assert type(data_row) == DataRow
    assert data_row.global_key == global_key


def test_get_data_row(data_row, client):
    assert client.get_data_row(data_row.uid)


def test_create_invalid_aws_data_row(dataset, client):
    with pytest.raises(InvalidQueryError) as exc:
        dataset.create_data_row(row_data="s3://labelbox-public-data/invalid")
    assert "s3" in exc.value.message

    with pytest.raises(InvalidQueryError) as exc:
        dataset.create_data_rows([{
            "row_data": "s3://labelbox-public-data/invalid"
        }])
    assert "s3" in exc.value.message


def test_lookup_data_rows(client, dataset):
    uid = str(uuid.uuid4())
    # 1 external id : 1 uid
    dr = dataset.create_data_row(row_data="123", external_id=uid)
    lookup = client.get_data_row_ids_for_external_ids([uid])
    assert len(lookup) == 1
    assert lookup[uid][0] == dr.uid
    # 2 external ids : 1 uid
    uid2 = str(uuid.uuid4())
    dr2 = dataset.create_data_row(row_data="123", external_id=uid2)
    lookup = client.get_data_row_ids_for_external_ids([uid, uid2])
    assert len(lookup) == 2
    assert all([len(x) == 1 for x in lookup.values()])
    assert lookup[uid][0] == dr.uid
    assert lookup[uid2][0] == dr2.uid
    # 1 external id : 2 uid
    dr3 = dataset.create_data_row(row_data="123", external_id=uid2)
    lookup = client.get_data_row_ids_for_external_ids([uid2])
    assert len(lookup) == 1
    assert len(lookup[uid2]) == 2
    assert lookup[uid2][0] == dr2.uid
    assert lookup[uid2][1] == dr3.uid
    # Empty args
    lookup = client.get_data_row_ids_for_external_ids([])
    assert len(lookup) == 0
    # Non matching
    lookup = client.get_data_row_ids_for_external_ids([str(uuid.uuid4())])
    assert len(lookup) == 0


def test_data_row_bulk_creation(dataset, rand_gen, image_url):
    client = dataset.client
    data_rows = []
    assert len(list(dataset.data_rows())) == 0

    try:
        payload = [
            {
                DataRow.row_data: image_url
            },
            {
                "row_data": image_url
            },
        ]
        with patch('labelbox.schema.dataset.UPSERT_CHUNK_SIZE_BYTES',
                   new=300):  # To make 2 chunks
            # Test creation using URL
            task = dataset.create_data_rows(payload, file_upload_thread_count=2)
        task.wait_till_done()
        assert task.has_errors() is False
        assert task.status == "COMPLETE"

        results = task.result
        assert len(results) == 2
        row_data = [result["row_data"] for result in results]
        assert row_data == [image_url, image_url]

        data_rows = list(dataset.data_rows())
        assert len(data_rows) == 2
        assert {data_row.row_data for data_row in data_rows} == {image_url}
        assert {data_row.global_key for data_row in data_rows} == {None}

        data_rows = list(dataset.data_rows(from_cursor=data_rows[0].uid))
        assert len(data_rows) == 1

    finally:
        for dr in data_rows:
            dr.delete()


@pytest.fixture
def local_image_file(image_url) -> NamedTemporaryFile:
    response = requests.get(image_url, stream=True)
    response.raise_for_status()

    with NamedTemporaryFile(delete=False) as f:
        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                f.write(chunk)

    yield f  # Return the path to the temp file

    os.remove(f.name)


def test_data_row_bulk_creation_from_file(dataset, local_image_file, image_url):
    with patch('labelbox.schema.dataset.UPSERT_CHUNK_SIZE_BYTES',
               new=500):  # Force chunking
        task = dataset.create_data_rows(
            [local_image_file.name, local_image_file.name])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        assert len(task.result) == 2
        assert task.has_errors() is False
        results = task.result
        row_data = [result["row_data"] for result in results]
        assert len(row_data) == 2


def test_data_row_bulk_creation_from_row_data_file_external_id(
        dataset, local_image_file, image_url):
    with patch('labelbox.schema.dataset.UPSERT_CHUNK_SIZE_BYTES',
               new=500):  # Force chunking
        task = dataset.create_data_rows([{
            "row_data": local_image_file.name,
            'external_id': 'some_name'
        }, {
            "row_data": image_url,
            'external_id': 'some_name2'
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        assert len(task.result) == 2
        assert task.has_errors() is False
        results = task.result
        row_data = [result["row_data"] for result in results]
        assert len(row_data) == 2
        assert image_url in row_data


def test_data_row_bulk_creation_from_row_data_file(dataset, rand_gen,
                                                   local_image_file, image_url):
    with patch('labelbox.schema.dataset.UPSERT_CHUNK_SIZE_BYTES',
               new=500):  # Force chunking
        task = dataset.create_data_rows([{
            "row_data": local_image_file.name
        }, {
            "row_data": local_image_file.name
        }])
        task.wait_till_done()
        assert task.status == "COMPLETE"
        assert len(task.result) == 2
        assert task.has_errors() is False
        results = task.result
        row_data = [result["row_data"] for result in results]
        assert len(row_data) == 2


@pytest.mark.slow
def test_data_row_large_bulk_creation(dataset, image_url):
    # Do a longer task and expect it not to be complete immediately
    n_urls = 1000
    n_local = 250
    with NamedTemporaryFile() as fp:
        fp.write("Test data".encode())
        fp.flush()
        task = dataset.create_data_rows([{
            DataRow.row_data: image_url
        }] * n_urls + [fp.name] * n_local)
    task.wait_till_done()
    assert task.status == "COMPLETE"
    assert len(list(dataset.data_rows())) == n_local + n_urls


def test_data_row_single_creation(dataset, rand_gen, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0

    data_row = dataset.create_data_row(row_data=image_url)
    assert len(list(dataset.data_rows())) == 1
    assert data_row.dataset() == dataset
    assert data_row.created_by() == client.get_user()
    assert data_row.organization() == client.get_organization()
    assert requests.get(image_url).content == \
        requests.get(data_row.row_data).content
    assert data_row.media_attributes is not None
    assert data_row.global_key is None

    with NamedTemporaryFile() as fp:
        data = rand_gen(str).encode()
        fp.write(data)
        fp.flush()
        data_row_2 = dataset.create_data_row(row_data=fp.name)
        assert len(list(dataset.data_rows())) == 2
        assert requests.get(data_row_2.row_data).content == data


def test_create_data_row_with_dict(dataset, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0
    dr = {"row_data": image_url}
    data_row = dataset.create_data_row(dr)
    assert len(list(dataset.data_rows())) == 1
    assert data_row.dataset() == dataset
    assert data_row.created_by() == client.get_user()
    assert data_row.organization() == client.get_organization()
    assert requests.get(image_url).content == \
        requests.get(data_row.row_data).content
    assert data_row.media_attributes is not None


def test_create_data_row_with_dict_containing_field(dataset, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0
    dr = {DataRow.row_data: image_url}
    data_row = dataset.create_data_row(dr)
    assert len(list(dataset.data_rows())) == 1
    assert data_row.dataset() == dataset
    assert data_row.created_by() == client.get_user()
    assert data_row.organization() == client.get_organization()
    assert requests.get(image_url).content == \
        requests.get(data_row.row_data).content
    assert data_row.media_attributes is not None


def test_create_data_row_with_dict_unpacked(dataset, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0
    dr = {"row_data": image_url}
    data_row = dataset.create_data_row(**dr)
    assert len(list(dataset.data_rows())) == 1
    assert data_row.dataset() == dataset
    assert data_row.created_by() == client.get_user()
    assert data_row.organization() == client.get_organization()
    assert requests.get(image_url).content == \
        requests.get(data_row.row_data).content
    assert data_row.media_attributes is not None


def test_create_data_row_with_invalid_input(dataset, image_url):
    with pytest.raises(ResourceCreationError) as exc:
        dataset.create_data_row("asdf")


def test_create_data_row_with_metadata(mdo, dataset, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0

    data_row = dataset.create_data_row(row_data=image_url,
                                       metadata_fields=make_metadata_fields())

    assert len(list(dataset.data_rows())) == 1
    assert data_row.dataset() == dataset
    assert data_row.created_by() == client.get_user()
    assert data_row.organization() == client.get_organization()
    assert requests.get(image_url).content == \
        requests.get(data_row.row_data).content
    assert data_row.media_attributes is not None
    metadata_fields = data_row.metadata_fields
    metadata = data_row.metadata
    assert len(metadata_fields) == 3
    assert len(metadata) == 3
    assert [m["schemaId"] for m in metadata_fields
           ].sort() == EXPECTED_METADATA_SCHEMA_IDS
    for m in metadata:
        assert mdo._parse_upsert(m)


def test_create_data_row_with_metadata_dict(mdo, dataset, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0

    data_row = dataset.create_data_row(
        row_data=image_url, metadata_fields=make_metadata_fields_dict())

    assert len(list(dataset.data_rows())) == 1
    assert data_row.dataset() == dataset
    assert data_row.created_by() == client.get_user()
    assert data_row.organization() == client.get_organization()
    assert requests.get(image_url).content == \
        requests.get(data_row.row_data).content
    assert data_row.media_attributes is not None
    metadata_fields = data_row.metadata_fields
    metadata = data_row.metadata
    assert len(metadata_fields) == 3
    assert len(metadata) == 3
    assert [m["schemaId"] for m in metadata_fields
           ].sort() == EXPECTED_METADATA_SCHEMA_IDS
    for m in metadata:
        assert mdo._parse_upsert(m)


def test_create_data_row_with_invalid_metadata(dataset, image_url):
    fields = make_metadata_fields()
    # make the payload invalid by providing the same schema id more than once
    fields.append(
        DataRowMetadataField(schema_id=TEXT_SCHEMA_ID, value='some msg'))

    with pytest.raises(ResourceCreationError):
        dataset.create_data_row(row_data=image_url, metadata_fields=fields)


def test_create_data_rows_with_metadata(mdo, dataset, image_url):
    client = dataset.client
    assert len(list(dataset.data_rows())) == 0

    task = dataset.create_data_rows([
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row1",
            DataRow.metadata_fields: make_metadata_fields()
        },
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row2",
            "metadata_fields": make_metadata_fields()
        },
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row3",
            DataRow.metadata_fields: make_metadata_fields_dict()
        },
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row4",
            "metadata_fields": make_metadata_fields_dict()
        },
    ])
    task.wait_till_done()

    assert len(list(dataset.data_rows())) == 4
    for r in ["row1", "row2", "row3", "row4"]:
        row = list(dataset.data_rows(where=DataRow.external_id == r))[0]
        assert row.dataset() == dataset
        assert row.created_by() == client.get_user()
        assert row.organization() == client.get_organization()
        assert requests.get(image_url).content == \
            requests.get(row.row_data).content
        assert row.media_attributes is not None

        metadata_fields = row.metadata_fields
        metadata = row.metadata
        assert len(metadata_fields) == 3
        assert len(metadata) == 3
        assert [m["schemaId"] for m in metadata_fields
               ].sort() == EXPECTED_METADATA_SCHEMA_IDS
        for m in metadata:
            assert mdo._parse_upsert(m)


@pytest.mark.parametrize("test_function,metadata_obj_type",
                         [("create_data_rows", "class"),
                          ("create_data_rows", "dict"),
                          ("create_data_rows_sync", "class"),
                          ("create_data_rows_sync", "dict"),
                          ("create_data_row", "class"),
                          ("create_data_row", "dict")])
def test_create_data_rows_with_named_metadata_field_class(
        test_function, metadata_obj_type, mdo, dataset, image_url):

    row_with_metadata_field = {
        DataRow.row_data:
            image_url,
        DataRow.external_id:
            "row1",
        DataRow.metadata_fields: [
            DataRowMetadataField(name='split', value='test'),
            DataRowMetadataField(name=CUSTOM_TEXT_SCHEMA_NAME, value='hello')
        ]
    }

    row_with_metadata_dict = {
        DataRow.row_data:
            image_url,
        DataRow.external_id:
            "row2",
        "metadata_fields": [
            {
                'name': 'split',
                'value': 'test'
            },
            {
                'name': CUSTOM_TEXT_SCHEMA_NAME,
                'value': 'hello'
            },
        ]
    }

    assert len(list(dataset.data_rows())) == 0

    METADATA_FIELDS = {
        "class": row_with_metadata_field,
        "dict": row_with_metadata_dict
    }

    def create_data_row(data_rows):
        dataset.create_data_row(data_rows[0])

    CREATION_FUNCTION = {
        "create_data_rows": dataset.create_data_rows,
        "create_data_rows_sync": dataset.create_data_rows_sync,
        "create_data_row": create_data_row
    }
    data_rows = [METADATA_FIELDS[metadata_obj_type]]
    function_to_test = CREATION_FUNCTION[test_function]
    task = function_to_test(data_rows)

    if isinstance(task, Task):
        task.wait_till_done()

    created_rows = list(dataset.data_rows())
    assert len(created_rows) == 1
    assert len(created_rows[0].metadata_fields) == 2
    assert len(created_rows[0].metadata) == 2

    metadata = created_rows[0].metadata
    assert metadata[0].schema_id == SPLIT_SCHEMA_ID
    assert metadata[0].name == 'test'
    assert metadata[0].value == mdo.reserved_by_name['split']['test'].uid
    assert metadata[1].name == CUSTOM_TEXT_SCHEMA_NAME
    assert metadata[1].value == 'hello'
    assert metadata[1].schema_id == mdo.custom_by_name[
        CUSTOM_TEXT_SCHEMA_NAME].uid


def test_create_data_rows_with_invalid_metadata(dataset, image_url):
    fields = make_metadata_fields()
    # make the payload invalid by providing the same schema id more than once
    fields.append(
        DataRowMetadataField(schema_id=TEXT_SCHEMA_ID, value="some msg"))

    task = dataset.create_data_rows([{
        DataRow.row_data: image_url,
        DataRow.metadata_fields: fields
    }])
    task.wait_till_done(timeout_seconds=60)

    assert task.status == "COMPLETE"
    assert len(task.failed_data_rows) == 1
    assert f"A schemaId can only be specified once per DataRow : [{TEXT_SCHEMA_ID}]" in task.failed_data_rows[
        0]["message"]


def test_create_data_rows_with_metadata_missing_value(dataset, image_url):
    fields = make_metadata_fields()
    fields.append({"schemaId": "some schema id"})

    with pytest.raises(ValueError) as exc:
        dataset.create_data_rows([
            {
                DataRow.row_data: image_url,
                DataRow.external_id: "row1",
                DataRow.metadata_fields: fields
            },
        ])


def test_create_data_rows_with_metadata_missing_schema_id(dataset, image_url):
    fields = make_metadata_fields()
    fields.append({"value": "some value"})

    with pytest.raises(ValueError) as exc:
        dataset.create_data_rows([
            {
                DataRow.row_data: image_url,
                DataRow.external_id: "row1",
                DataRow.metadata_fields: fields
            },
        ])


def test_create_data_rows_with_metadata_wrong_type(dataset, image_url):
    fields = make_metadata_fields()
    fields.append("Neither DataRowMetadataField or dict")

    with pytest.raises(ValueError) as exc:
        task = dataset.create_data_rows([
            {
                DataRow.row_data: image_url,
                DataRow.external_id: "row1",
                DataRow.metadata_fields: fields
            },
        ])


def test_data_row_update_missing_or_empty_required_fields(
        dataset, rand_gen, image_url):
    external_id = rand_gen(str)
    data_row = dataset.create_data_row(row_data=image_url,
                                       external_id=external_id)
    with pytest.raises(ValueError):
        data_row.update(row_data="")
    with pytest.raises(ValueError):
        data_row.update(row_data={})
    with pytest.raises(ValueError):
        data_row.update(external_id="")
    with pytest.raises(ValueError):
        data_row.update(global_key="")
    with pytest.raises(ValueError):
        data_row.update()


def test_data_row_update(client, dataset, rand_gen, image_url,
                         wait_for_data_row_processing):
    external_id = rand_gen(str)
    data_row = dataset.create_data_row(row_data=image_url,
                                       external_id=external_id)
    assert data_row.external_id == external_id

    external_id_2 = rand_gen(str)
    data_row.update(external_id=external_id_2)
    assert data_row.external_id == external_id_2

    in_line_content = "123"
    data_row.update(row_data=in_line_content)
    assert data_row.row_data == in_line_content

    data_row.update(row_data=image_url)
    data_row = wait_for_data_row_processing(client, data_row)
    assert data_row.row_data == image_url

    # tileLayer becomes a media attribute
    pdf_url = "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483.pdf"
    tileLayerUrl = "https://storage.googleapis.com/labelbox-datasets/arxiv-pdf/data/99-word-token-pdfs/0801.3483-lb-textlayer.json"
    data_row.update(row_data={'pdfUrl': pdf_url, "tileLayerUrl": tileLayerUrl})
    custom_check = lambda data_row: data_row.row_data and 'pdfUrl' not in data_row.row_data
    data_row = wait_for_data_row_processing(client,
                                            data_row,
                                            custom_check=custom_check)
    assert data_row.row_data == pdf_url


def test_data_row_filtering_sorting(dataset, image_url):
    task = dataset.create_data_rows([
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row1"
        },
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row2"
        },
    ])
    task.wait_till_done()

    # Test filtering
    row1 = list(dataset.data_rows(where=DataRow.external_id == "row1"))
    assert len(row1) == 1
    row1 = dataset.data_rows_for_external_id("row1")
    assert len(row1) == 1
    row1 = row1[0]
    assert row1.external_id == "row1"
    row2 = list(dataset.data_rows(where=DataRow.external_id == "row2"))
    assert len(row2) == 1
    row2 = dataset.data_rows_for_external_id("row2")
    assert len(row2) == 1
    row2 = row2[0]
    assert row2.external_id == "row2"


@pytest.fixture
def create_datarows_for_data_row_deletion(dataset, image_url):
    task = dataset.create_data_rows([{
        DataRow.row_data: image_url,
        DataRow.external_id: str(i)
    } for i in range(10)])
    task.wait_till_done()

    data_rows = list(dataset.data_rows())

    yield data_rows
    for dr in data_rows:
        dr.delete()


def test_data_row_deletion(dataset, create_datarows_for_data_row_deletion):
    create_datarows_for_data_row_deletion
    data_rows = list(dataset.data_rows())
    expected = set(map(str, range(10)))
    assert {dr.external_id for dr in data_rows} == expected

    for dr in data_rows:
        if dr.external_id in "37":
            dr.delete()
    expected -= set("37")

    data_rows = list(dataset.data_rows())
    assert {dr.external_id for dr in data_rows} == expected

    DataRow.bulk_delete([dr for dr in data_rows if dr.external_id in "2458"])
    expected -= set("2458")

    data_rows = list(dataset.data_rows())
    assert {dr.external_id for dr in data_rows} == expected


def test_data_row_iteration(dataset, image_url) -> None:
    task = dataset.create_data_rows([
        {
            DataRow.row_data: image_url
        },
        {
            "row_data": image_url
        },
    ])
    task.wait_till_done()
    assert next(dataset.data_rows())


def test_data_row_attachments(dataset, image_url):
    attachments = [("IMAGE", image_url, "attachment image"),
                   ("RAW_TEXT", "test-text", None),
                   ("IMAGE_OVERLAY", image_url, "Overlay"),
                   ("HTML", image_url, None)]
    task = dataset.create_data_rows([{
        "row_data":
            image_url,
        "external_id":
            "test-id",
        "attachments": [{
            "type": attachment_type,
            "value": attachment_value,
            "name": attachment_name
        }]
    } for attachment_type, attachment_value, attachment_name in attachments])

    task.wait_till_done()
    assert task.status == "COMPLETE"
    data_rows = list(dataset.data_rows())
    assert len(data_rows) == len(attachments)
    for data_row in data_rows:
        assert len(list(data_row.attachments())) == 1
        assert data_row.external_id == "test-id"

    with pytest.raises(ValueError) as exc:
        task = dataset.create_data_rows([{
            "row_data": image_url,
            "external_id": "test-id",
            "attachments": [{
                "type": "INVALID",
                "value": "123"
            }]
        }])


def test_create_data_rows_sync_attachments(dataset, image_url):
    attachments = [("IMAGE", image_url, "image URL"),
                   ("RAW_TEXT", "test-text", None),
                   ("IMAGE_OVERLAY", image_url, "Overlay"),
                   ("HTML", image_url, None)]
    attachments_per_data_row = 3
    dataset.create_data_rows_sync([{
        "row_data":
            image_url,
        "external_id":
            "test-id",
        "attachments": [{
            "type": attachment_type,
            "value": attachment_value,
            "name": attachment_name
        } for _ in range(attachments_per_data_row)]
    } for attachment_type, attachment_value, attachment_name in attachments])
    data_rows = list(dataset.data_rows())
    assert len(data_rows) == len(attachments)
    for data_row in data_rows:
        assert len(list(data_row.attachments())) == attachments_per_data_row


def test_create_data_rows_sync_mixed_upload(dataset, image_url):
    n_local = 100
    n_urls = 100
    with NamedTemporaryFile() as fp:
        fp.write("Test data".encode())
        fp.flush()
        dataset.create_data_rows_sync([{
            DataRow.row_data: image_url
        }] * n_urls + [fp.name] * n_local)
    assert len(list(dataset.data_rows())) == n_local + n_urls


def test_create_data_row_attachment(data_row):
    att = data_row.create_attachment("IMAGE", "https://example.com/image.jpg",
                                     "name")
    assert att.attachment_type == "IMAGE"
    assert att.attachment_value == "https://example.com/image.jpg"
    assert att.attachment_name == "name"


def test_create_data_row_attachment_invalid_type(data_row):
    with pytest.raises(ValueError):
        data_row.create_attachment("SOME_TYPE", "value", "name")


def test_create_data_row_attachment_invalid_value(data_row):
    with pytest.raises(ValueError):
        data_row.create_attachment("IMAGE", "", "name")
    with pytest.raises(ValueError):
        data_row.create_attachment("IMAGE", None, "name")


def test_delete_data_row_attachment(data_row, image_url):
    attachments = []

    # Anonymous attachment
    to_attach = [("IMAGE", image_url), ("RAW_TEXT", "test-text"),
                 ("IMAGE_OVERLAY", image_url), ("HTML", image_url)]
    for attachment_type, attachment_value in to_attach:
        attachments.append(
            data_row.create_attachment(attachment_type, attachment_value))

    # Attachment with a name
    to_attach = [("IMAGE", image_url, "Att. Image"),
                 ("RAW_TEXT", "test-text", "Att. Text"),
                 ("IMAGE_OVERLAY", image_url, "Image Overlay"),
                 ("HTML", image_url, "Att. HTML")]
    for attachment_type, attachment_value, attachment_name in to_attach:
        attachments.append(
            data_row.create_attachment(attachment_type, attachment_value,
                                       attachment_name))

    for attachment in attachments:
        attachment.delete()

    assert len(list(data_row.attachments())) == 0


def test_update_data_row_attachment(data_row, image_url):
    attachment: AssetAttachment = data_row.create_attachment(
        "RAW_TEXT", "value", "name")
    assert attachment is not None
    attachment.update(name="updated name", type="IMAGE", value=image_url)
    assert attachment.attachment_name == "updated name"
    assert attachment.attachment_type == "IMAGE"
    assert attachment.attachment_value == image_url


def test_update_data_row_attachment_invalid_type(data_row):
    attachment: AssetAttachment = data_row.create_attachment(
        "RAW_TEXT", "value", "name")
    assert attachment is not None
    with pytest.raises(ValueError):
        attachment.update(name="updated name", type="INVALID", value="value")


def test_update_data_row_attachment_invalid_value(data_row):
    attachment: AssetAttachment = data_row.create_attachment(
        "RAW_TEXT", "value", "name")
    assert attachment is not None
    with pytest.raises(ValueError):
        attachment.update(name="updated name", type="IMAGE", value="")


def test_does_not_update_not_provided_attachment_fields(data_row):
    attachment: AssetAttachment = data_row.create_attachment(
        "RAW_TEXT", "value", "name")
    assert attachment is not None
    attachment.update(value=None, name="name")
    assert attachment.attachment_value == "value"
    attachment.update(name=None, value="value")
    assert attachment.attachment_name == "name"
    attachment.update(type=None, name="name")
    assert attachment.attachment_type == "RAW_TEXT"


def test_create_data_rows_result(client, dataset, image_url):
    task = dataset.create_data_rows([
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row1",
        },
        {
            DataRow.row_data: image_url,
            DataRow.external_id: "row1",
        },
    ])
    task.wait_till_done()
    assert task.errors is None
    for result in task.result:
        client.get_data_row(result['id'])


def test_create_data_rows_local_file(dataset, sample_image):
    task = dataset.create_data_rows([{
        DataRow.row_data: sample_image,
        DataRow.metadata_fields: make_metadata_fields()
    }])
    task.wait_till_done()
    assert task.status == "COMPLETE"
    data_row = list(dataset.data_rows())[0]
    assert data_row.external_id == "tests/integration/media/sample_image.jpg"
    assert len(data_row.metadata_fields) == 3


def test_data_row_with_global_key(dataset, sample_image):
    global_key = str(uuid.uuid4())
    row = dataset.create_data_row({
        DataRow.row_data: sample_image,
        DataRow.global_key: global_key
    })

    assert row.global_key == global_key


def test_data_row_bulk_creation_with_unique_global_keys(dataset, sample_image):
    global_key_1 = str(uuid.uuid4())
    global_key_2 = str(uuid.uuid4())
    global_key_3 = str(uuid.uuid4())

    task = dataset.create_data_rows([
        {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_1
        },
        {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_2
        },
        {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_3
        },
    ])

    task.wait_till_done()
    assert {row.global_key for row in dataset.data_rows()
           } == {global_key_1, global_key_2, global_key_3}


def test_data_row_bulk_creation_with_same_global_keys(dataset, sample_image,
                                                      snapshot):
    global_key_1 = str(uuid.uuid4())
    task = dataset.create_data_rows([{
        DataRow.row_data: sample_image,
        DataRow.global_key: global_key_1
    }, {
        DataRow.row_data: sample_image,
        DataRow.global_key: global_key_1
    }])

    task.wait_till_done()

    assert task.status == "COMPLETE"
    assert type(task.failed_data_rows) is list
    assert len(task.failed_data_rows) == 1
    assert type(task.created_data_rows) is list
    assert len(task.created_data_rows) == 1
    assert task.failed_data_rows[0][
        'message'] == f"Duplicate global key: '{global_key_1}'"
    assert task.failed_data_rows[0]['failedDataRows'][0][
        'externalId'] == sample_image
    assert task.created_data_rows[0]['external_id'] == sample_image
    assert task.created_data_rows[0]['global_key'] == global_key_1

    assert len(task.errors) == 1
    assert task.has_errors() is True

    all_results = task.result
    assert len(all_results) == 1


def test_data_row_delete_and_create_with_same_global_key(
        client, dataset, sample_image):
    global_key_1 = str(uuid.uuid4())
    data_row_payload = {
        DataRow.row_data: sample_image,
        DataRow.global_key: global_key_1
    }

    # should successfully insert new datarow
    task = dataset.create_data_rows([data_row_payload])
    task.wait_till_done()

    assert task.status == "COMPLETE"
    assert task.result[0]['global_key'] == global_key_1

    new_data_row_id = task.result[0]['id']

    # same payload should fail due to duplicated global key
    task = dataset.create_data_rows([data_row_payload])
    task.wait_till_done()

    assert task.status == "COMPLETE"
    assert len(task.failed_data_rows) == 1
    assert task.failed_data_rows[0][
        'message'] == f"Duplicate global key: '{global_key_1}'"

    # delete datarow
    client.get_data_row(new_data_row_id).delete()

    # should successfully insert new datarow now
    task = dataset.create_data_rows([data_row_payload])
    task.wait_till_done()

    assert task.status == "COMPLETE"
    assert task.result[0]['global_key'] == global_key_1


def test_data_row_bulk_creation_sync_with_unique_global_keys(
        dataset, sample_image):
    global_key_1 = str(uuid.uuid4())
    global_key_2 = str(uuid.uuid4())
    global_key_3 = str(uuid.uuid4())

    dataset.create_data_rows_sync([
        {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_1
        },
        {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_2
        },
        {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_3
        },
    ])

    assert {row.global_key for row in dataset.data_rows()
           } == {global_key_1, global_key_2, global_key_3}


def test_data_row_bulk_creation_sync_with_same_global_keys(
        dataset, sample_image):
    global_key_1 = str(uuid.uuid4())

    with pytest.raises(ResourceCreationError) as exc_info:
        dataset.create_data_rows_sync([{
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_1
        }, {
            DataRow.row_data: sample_image,
            DataRow.global_key: global_key_1
        }])

    assert len(list(dataset.data_rows())) == 1
    assert list(dataset.data_rows())[0].global_key == global_key_1
    assert "Duplicate global key" in str(exc_info.value)
    assert exc_info.value.args[1]  # task id


@pytest.fixture
def conversational_data_rows(dataset, conversational_content):
    examples = [
        {
            **conversational_content, 'media_type':
                MediaType.Conversational.value
        },
        conversational_content,
        {
            "conversationalData": conversational_content['row_data']['messages']
        }  # Old way to check for backwards compatibility
    ]
    task = dataset.create_data_rows(examples)
    task.wait_till_done()
    assert task.status == "COMPLETE"

    data_rows = list(dataset.data_rows())

    yield data_rows
    for dr in data_rows:
        dr.delete()


def test_create_conversational_text(conversational_data_rows,
                                    conversational_content):
    data_rows = conversational_data_rows
    for data_row in data_rows:
        assert json.loads(
            data_row.row_data) == conversational_content['row_data']


def test_invalid_media_type(dataset, conversational_content):
    for _, __ in [["Found invalid contents for media type: 'IMAGE'", 'IMAGE'],
                  [
                      "Found invalid media type: 'totallyinvalid'",
                      'totallyinvalid'
                  ]]:
        # TODO: What error kind should this be? It looks like for global key we are
        # using malformed query. But for invalid contents in FileUploads we use InvalidQueryError
        with pytest.raises(ResourceCreationError):
            dataset.create_data_rows_sync([{
                **conversational_content, 'media_type': 'IMAGE'
            }])


def test_create_tiled_layer(dataset, tile_content):
    examples = [
        {
            **tile_content, 'media_type': 'TMS_GEO'
        },
        tile_content,
    ]
    dataset.create_data_rows_sync(examples)
    data_rows = list(dataset.data_rows())
    assert len(data_rows) == len(examples)
    for data_row in data_rows:
        assert json.loads(data_row.row_data) == tile_content['row_data']


def test_create_data_row_with_attachments(dataset):
    attachment_value = 'attachment value'
    dr = dataset.create_data_row(row_data="123",
                                 attachments=[{
                                     'type': 'RAW_TEXT',
                                     'value': attachment_value
                                 }])
    attachments = list(dr.attachments())
    assert len(attachments) == 1


def test_create_data_row_with_media_type(dataset, image_url):
    with pytest.raises(ResourceCreationError) as exc:
        dr = dataset.create_data_row(
            row_data={'invalid_object': 'invalid_value'}, media_type="IMAGE")

    assert "Expected type image/*, detected: application/json" in str(exc.value)

    dataset.create_data_row(row_data=image_url, media_type="IMAGE")



import copy
import logging
import uuid
from collections import defaultdict, deque
from typing import Any, Deque, Dict, Generator, Iterable, List, Set, Union

from labelbox.data.annotation_types.annotation import ObjectAnnotation
from labelbox.data.annotation_types.classification.classification import (
    ClassificationAnnotation,)
from labelbox.data.annotation_types.metrics.confusion_matrix import (
    ConfusionMatrixMetric,)
from labelbox.data.annotation_types.metrics.scalar import ScalarMetric
from labelbox.data.annotation_types.video import VideoMaskAnnotation

from ...annotation_types.collection import LabelCollection, LabelGenerator
from ...annotation_types.relationship import RelationshipAnnotation
from .label import NDLabel

logger = logging.getLogger(__name__)

IGNORE_IF_NONE = ["page", "unit", "messageId"]


class NDJsonConverter:

    @staticmethod
    def deserialize(json_data: Iterable[Dict[str, Any]]) -> LabelGenerator:
        """
        Converts ndjson data (prediction import format) into the common labelbox format.

        Args:
            json_data: An iterable representing the ndjson data
        Returns:
            LabelGenerator containing the ndjson data.
        """
        data = NDLabel(**{"annotations": json_data})
        res = data.to_common()
        return res

    @staticmethod
    def serialize(
            labels: LabelCollection) -> Generator[Dict[str, Any], None, None]:
        """
        Converts a labelbox common object to the labelbox ndjson format (prediction import format)

        Note that this function might fail for objects that are not supported by mal.
        Not all edge cases are handling by custom exceptions, if you get a cryptic pydantic error message it is probably due to this.
        We will continue to improve the error messages and add helper functions to deal with this.

        Args:
            labels: Either a list of Label objects or a LabelGenerator
        Returns:
            A generator for accessing the ndjson representation of the data
        """
        used_uuids: Set[uuid.UUID] = set()

        relationship_uuids: Dict[uuid.UUID,
                                 Deque[uuid.UUID]] = defaultdict(deque)

        # UUIDs are private properties used to enhance UX when defining relationships.
        # They are created for all annotations, but only utilized for relationships.
        # To avoid overwriting, UUIDs must be unique across labels.
        # Non-relationship annotation UUIDs are regenerated when they are reused.
        # For relationship annotations, during first pass, we update the UUIDs of the source and target annotations.
        # During the second pass, we update the UUIDs of the annotations referenced by the relationship annotations.
        for label in labels:
            uuid_safe_annotations: List[Union[
                ClassificationAnnotation,
                ObjectAnnotation,
                VideoMaskAnnotation,
                ScalarMetric,
                ConfusionMatrixMetric,
                RelationshipAnnotation,
            ]] = []
            # First pass to get all RelatiohnshipAnnotaitons
            # and update the UUIDs of the source and target annotations
            for annotation in label.annotations:
                if isinstance(annotation, RelationshipAnnotation):
                    annotation = copy.deepcopy(annotation)
                    new_source_uuid = uuid.uuid4()
                    new_target_uuid = uuid.uuid4()
                    relationship_uuids[annotation.value.source._uuid].append(
                        new_source_uuid)
                    relationship_uuids[annotation.value.target._uuid].append(
                        new_target_uuid)
                    annotation.value.source._uuid = new_source_uuid
                    annotation.value.target._uuid = new_target_uuid
                    if annotation._uuid in used_uuids:
                        annotation._uuid = uuid.uuid4()
                    used_uuids.add(annotation._uuid)
                    uuid_safe_annotations.append(annotation)
            # Second pass to update UUIDs for annotations referenced by RelationshipAnnotations
            for annotation in label.annotations:
                if (not isinstance(annotation, RelationshipAnnotation) and
                        hasattr(annotation, "_uuid")):
                    annotation = copy.deepcopy(annotation)
                    next_uuids = relationship_uuids[annotation._uuid]
                    if len(next_uuids) > 0:
                        annotation._uuid = next_uuids.popleft()

                    if annotation._uuid in used_uuids:
                        annotation._uuid = uuid.uuid4()
                    used_uuids.add(annotation._uuid)
                    uuid_safe_annotations.append(annotation)
                else:
                    if not isinstance(annotation, RelationshipAnnotation):
                        uuid_safe_annotations.append(annotation)
            label.annotations = uuid_safe_annotations
            for example in NDLabel.from_common([label]):
                annotation_uuid = getattr(example, "uuid", None)

                res = example.dict(
                    by_alias=True,
                    exclude={"uuid"} if annotation_uuid == "None" else None,
                )
                for k, v in list(res.items()):
                    if k in IGNORE_IF_NONE and v is None:
                        del res[k]
                yield res



from enum import Enum


class MediaType(Enum):
    Audio = "AUDIO"
    Conversational = "CONVERSATIONAL"
    Dicom = "DICOM"
    Document = "PDF"
    Geospatial_Tile = "TMS_GEO"
    Html = "HTML"
    Image = "IMAGE"
    LLMPromptCreation = "LLM_PROMPT_CREATION"
    LLMPromptResponseCreation = "LLM_PROMPT_RESPONSE_CREATION"
    Pdf = "PDF"
    Simple_Tile = "TMS_SIMPLE"
    Text = "TEXT"
    Tms_Geo = "TMS_GEO"
    Tms_Simple = "TMS_SIMPLE"
    Unknown = "UNKNOWN"
    Unsupported = "UNSUPPORTED"
    Video = "VIDEO"
    Point_Cloud = "POINT_CLOUD"
    LLM = "LLM"

    @classmethod
    def _missing_(cls, name):
        """Handle missing null data types for projects
            created without setting allowedMediaType
            Handle upper case names for compatibility with
            the GraphQL"""

        if name is None:
            return cls.Unknown

        for member in cls.__members__:
            if member.name == name.upper():
                return member

    @classmethod
    def is_supported(cls, value):
        return isinstance(value,
                          cls) and value not in [cls.Unknown, cls.Unsupported]

    @classmethod
    def get_supported_members(cls):
        return [
            item for item in cls.__members__
            if item not in ["Unknown", "Unsupported"]
        ]


def get_media_type_validation_error(media_type):
    return TypeError(f"{media_type} is not a valid media type. Use"
                     f" any of {MediaType.get_supported_members()}"
                     " from MediaType. Example: MediaType.Image.")



from typing import Any, Tuple, List, Union
from pathlib import Path
from collections import defaultdict
import warnings

from ...annotation_types.relationship import RelationshipAnnotation
from ...annotation_types.metrics.confusion_matrix import ConfusionMatrixMetric
from ...annotation_types.metrics.scalar import ScalarMetric
from ...annotation_types.video import VideoMaskAnnotation
from ...annotation_types.annotation import ObjectAnnotation
from ...annotation_types.classification.classification import ClassificationAnnotation

from .... import pydantic_compat
import numpy as np

from .path import PathSerializerMixin


def rle_decoding(rle_arr: List[int], w: int, h: int) -> np.ndarray:
    indices = []
    for idx, cnt in zip(rle_arr[0::2], rle_arr[1::2]):
        indices.extend(list(range(idx - 1,
                                  idx + cnt - 1)))  # RLE is 1-based index
    mask = np.zeros(h * w, dtype=np.uint8)
    mask[indices] = 1
    return mask.reshape((w, h)).T


def get_annotation_lookup(annotations):
    """Get annotations from Label.annotations objects

    Args:
        annotations (Label.annotations): Annotations attached to labelbox Label object used as private method
    """
    annotation_lookup = defaultdict(list)
    for annotation in annotations:
        # Provide a default value of None if the attribute doesn't exist
        attribute_value = getattr(annotation, 'image_id', None) or getattr(annotation, 'name', None)
        annotation_lookup[attribute_value].append(annotation)
    return annotation_lookup 


class SegmentInfo(pydantic_compat.BaseModel):
    id: int
    category_id: int
    area: int
    bbox: Tuple[float, float, float, float]  #[x,y,w,h],
    iscrowd: int = 0


class RLE(pydantic_compat.BaseModel):
    counts: List[int]
    size: Tuple[int, int]  # h,w or w,h?


class COCOObjectAnnotation(pydantic_compat.BaseModel):
    # All segmentations for a particular class in an image...
    # So each image will have one of these for each class present in the image..
    # Annotations only exist if there is data..
    id: int
    image_id: int
    category_id: int
    segmentation: Union[RLE, List[List[float]]]  # [[x1,y1,x2,y2,x3,y3...]]
    area: float
    bbox: Tuple[float, float, float, float]  #[x,y,w,h],
    iscrowd: int = 0


class PanopticAnnotation(PathSerializerMixin):
    # One to one relationship between image and panoptic annotation
    image_id: int
    file_name: Path
    segments_info: List[SegmentInfo]



import pytest

from labelbox.schema.queue_mode import QueueMode


def test_project_dataset(client, rand_gen):
    with pytest.raises(
            ValueError,
            match=
            "Dataset queue mode is deprecated. Please prefer Batch queue mode."
    ):
        client.create_project(
            name=rand_gen(str),
            queue_mode=QueueMode.Dataset,
        )


def test_project_auto_audit_parameters(client, rand_gen):
    with pytest.raises(
            ValueError,
            match=
            "quality_mode must be set instead of auto_audit_percentage or auto_audit_number_of_labels."
    ):
        client.create_project(name=rand_gen(str), auto_audit_percentage=0.5)

    with pytest.raises(
            ValueError,
            match=
            "quality_mode must be set instead of auto_audit_percentage or auto_audit_number_of_labels."
    ):
        client.create_project(name=rand_gen(str), auto_audit_number_of_labels=2)


def test_project_name_parameter(client, rand_gen):
    with pytest.raises(ValueError,
                       match="project name must be a valid string."):
        client.create_project()

    with pytest.raises(ValueError,
                       match="project name must be a valid string."):
        client.create_project(name="     ")



import pytest


@pytest.mark.parametrize('data_rows', [3], indirect=True)
def test_catalog_export_v2(client, export_v2_test_helpers, data_rows):
    datarow_filter_size = 2
    data_row_ids = [dr.uid for dr in data_rows]

    params = {"performance_details": False, "label_details": False}
    filters = {"data_row_ids": data_row_ids[:datarow_filter_size]}

    task_results = export_v2_test_helpers.run_catalog_export_v2_task(
        client, filters=filters, params=params)

    # only 2 datarows should be exported
    assert len(task_results) == datarow_filter_size
    # only filtered datarows should be exported
    assert set([dr['data_row']['id'] for dr in task_results
               ]) == set(data_row_ids[:datarow_filter_size])



import json

import pytest
from labelbox.data.annotation_types.geometry.polygon import Polygon
from labelbox.data.annotation_types.geometry.line import Line
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.rectangle import Rectangle

from labelbox.data.serialization.labelbox_v1.converter import LBV1Converter
from labelbox.schema.bulk_import_request import Bbox


@pytest.mark.parametrize(
    "file_path", ['tests/data/assets/labelbox_v1/tiled_image_export.json'])
def test_image(file_path):
    """Tests against both Simple and non-Simple tiled image export data. 
    index-0 is non-Simple, index-1 is Simple
    """
    with open(file_path, 'r') as f:
        payload = json.load(f)

        collection = LBV1Converter.deserialize(payload)
        collection_as_list = list(collection)

        assert len(collection_as_list) == 2

        non_simple_annotations = collection_as_list[0].annotations
        assert len(non_simple_annotations) == 6
        expected_shapes = [Polygon, Point, Point, Point, Line, Rectangle]
        for idx in range(len(non_simple_annotations)):
            assert isinstance(non_simple_annotations[idx].value,
                              expected_shapes[idx])
        assert non_simple_annotations[-1].value.start.x == -99.36567524971268
        assert non_simple_annotations[-1].value.start.y == 19.34717117508651
        assert non_simple_annotations[-1].value.end.x == -99.3649886680726
        assert non_simple_annotations[-1].value.end.y == 19.41999425190506

        simple_annotations = collection_as_list[1].annotations
        assert len(simple_annotations) == 8
        expected_shapes = [
            Polygon, Point, Point, Point, Point, Point, Line, Rectangle
        ]
        for idx in range(len(simple_annotations)):
            assert isinstance(simple_annotations[idx].value,
                              expected_shapes[idx])



from collections import defaultdict
from itertools import islice
import json
import os
import sys
import re
import time
import uuid
import requests
from types import SimpleNamespace
from typing import Type, List
from enum import Enum
from typing import Tuple

import pytest
import requests

from labelbox import Dataset, DataRow
from labelbox import LabelingFrontend
from labelbox import OntologyBuilder, Tool, Option, Classification, MediaType
from labelbox.orm import query
from labelbox.pagination import PaginatedCollection
from labelbox.schema.annotation_import import LabelImport
from labelbox.schema.catalog import Catalog
from labelbox.schema.enums import AnnotationImportState
from labelbox.schema.invite import Invite
from labelbox.schema.quality_mode import QualityMode
from labelbox.schema.queue_mode import QueueMode
from labelbox.schema.user import User
from labelbox import Client
from labelbox.schema.ontology_kind import OntologyKind


@pytest.fixture
def project_based_user(client, rand_gen):
    email = rand_gen(str)
    # Use old mutation because it doesn't require users to accept email invites
    query_str = """mutation MakeNewUserPyApi {
        addMembersToOrganization(
            data: {
                emails: ["%s@labelbox.com"],
                orgRoleId: "%s",
                projectRoles: []
            }
        ) {
        newUserId
        }
    }
    """ % (email, str(client.get_roles()['NONE'].uid))
    user_id = client.execute(
        query_str)['addMembersToOrganization'][0]['newUserId']
    assert user_id is not None, "Unable to add user with old mutation"
    user = client._get_single(User, user_id)
    yield user
    client.get_organization().remove_user(user)


@pytest.fixture
def project_pack(client):
    projects = [
        client.create_project(name=f"user-proj-{idx}",
                              queue_mode=QueueMode.Batch,
                              media_type=MediaType.Image) for idx in range(2)
    ]
    yield projects
    for proj in projects:
        proj.delete()


@pytest.fixture
def project_with_empty_ontology(project):
    editor = list(
        project.client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]
    empty_ontology = {"tools": [], "classifications": []}
    project.setup(editor, empty_ontology)
    yield project


@pytest.fixture
def configured_project(project_with_empty_ontology, initial_dataset, rand_gen,
                       image_url):
    dataset = initial_dataset
    data_row_id = dataset.create_data_row(row_data=image_url).uid
    project = project_with_empty_ontology

    batch = project.create_batch(
        rand_gen(str),
        [data_row_id],  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = [data_row_id]

    yield project

    batch.delete()


@pytest.fixture
def configured_project_with_complex_ontology(client, initial_dataset, rand_gen,
                                             image_url):
    project = client.create_project(name=rand_gen(str),
                                    queue_mode=QueueMode.Batch,
                                    media_type=MediaType.Image)
    dataset = initial_dataset
    data_row = dataset.create_data_row(row_data=image_url)
    data_row_ids = [data_row.uid]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids

    editor = list(
        project.client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]

    ontology = OntologyBuilder()
    tools = [
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class"),
        Tool(tool=Tool.Type.LINE, name="test-line-class"),
        Tool(tool=Tool.Type.POINT, name="test-point-class"),
        Tool(tool=Tool.Type.POLYGON, name="test-polygon-class"),
        Tool(tool=Tool.Type.NER, name="test-ner-class")
    ]

    options = [
        Option(value="first option answer"),
        Option(value="second option answer"),
        Option(value="third option answer")
    ]

    classifications = [
        Classification(class_type=Classification.Type.TEXT,
                       name="test-text-class"),
        Classification(class_type=Classification.Type.DROPDOWN,
                       name="test-dropdown-class",
                       options=options),
        Classification(class_type=Classification.Type.RADIO,
                       name="test-radio-class",
                       options=options),
        Classification(class_type=Classification.Type.CHECKLIST,
                       name="test-checklist-class",
                       options=options)
    ]

    for t in tools:
        for c in classifications:
            t.add_classification(c)
        ontology.add_tool(t)
    for c in classifications:
        ontology.add_classification(c)

    project.setup(editor, ontology.asdict())

    yield [project, data_row]
    project.delete()


@pytest.fixture
def ontology(client):
    ontology_builder = OntologyBuilder(
        tools=[
            Tool(tool=Tool.Type.BBOX, name="Box 1", color="#ff0000"),
            Tool(tool=Tool.Type.BBOX, name="Box 2", color="#ff0000")
        ],
        classifications=[
            Classification(name="Root Class",
                           class_type=Classification.Type.RADIO,
                           options=[
                               Option(value="1", label="Option 1"),
                               Option(value="2", label="Option 2")
                           ])
        ])
    ontology = client.create_ontology('Integration Test Ontology',
                                      ontology_builder.asdict(),
                                      MediaType.Image)
    yield ontology
    client.delete_unused_ontology(ontology.uid)


@pytest.fixture
def video_data(client, rand_gen, video_data_row, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    data_row = dataset.create_data_row(video_data_row)
    data_row = wait_for_data_row_processing(client, data_row)
    data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


def create_video_data_row(rand_gen):
    return {
        "row_data":
            "https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4",
        "global_key":
            f"https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4-{rand_gen(str)}",
        "media_type":
            "VIDEO",
    }


@pytest.fixture
def video_data_100_rows(client, rand_gen, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    for _ in range(100):
        data_row = dataset.create_data_row(create_video_data_row(rand_gen))
        data_row = wait_for_data_row_processing(client, data_row)
        data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


@pytest.fixture()
def video_data_row(rand_gen):
    return create_video_data_row(rand_gen)


class ExportV2Helpers:

    @classmethod
    def run_project_export_v2_task(cls,
                                   project,
                                   num_retries=5,
                                   task_name=None,
                                   filters={},
                                   params={}):
        task = None
        params = params if params else {
            "project_details": True,
            "performance_details": False,
            "data_row_details": True,
            "label_details": True
        }
        while (num_retries > 0):
            task = project.export_v2(task_name=task_name,
                                     filters=filters,
                                     params=params)
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break
        return task.result

    @classmethod
    def run_dataset_export_v2_task(cls,
                                   dataset,
                                   num_retries=5,
                                   task_name=None,
                                   filters={},
                                   params={}):
        task = None
        params = params if params else {
            "performance_details": False,
            "label_details": True
        }
        while (num_retries > 0):
            task = dataset.export_v2(task_name=task_name,
                                     filters=filters,
                                     params=params)
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result

    @classmethod
    def run_catalog_export_v2_task(cls,
                                   client,
                                   num_retries=5,
                                   task_name=None,
                                   filters={},
                                   params={}):
        task = None
        params = params if params else {
            "performance_details": False,
            "label_details": True
        }
        catalog = client.get_catalog()
        while (num_retries > 0):

            task = catalog.export_v2(task_name=task_name,
                                     filters=filters,
                                     params=params)
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result


@pytest.fixture
def export_v2_test_helpers() -> Type[ExportV2Helpers]:
    return ExportV2Helpers()


@pytest.fixture
def big_dataset_data_row_ids(big_dataset: Dataset):
    export_task = big_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    yield [dr.json["data_row"]["id"] for dr in stream]


@pytest.fixture(scope='function')
def dataset_with_invalid_data_rows(unique_dataset: Dataset,
                                   upload_invalid_data_rows_for_dataset):
    upload_invalid_data_rows_for_dataset(unique_dataset)

    yield unique_dataset


@pytest.fixture
def upload_invalid_data_rows_for_dataset():

    def _upload_invalid_data_rows_for_dataset(dataset: Dataset):
        task = dataset.create_data_rows([
            {
                "row_data": 'gs://invalid-bucket/example.png',  # forbidden
                "external_id": "image-without-access.jpg"
            },
        ] * 2)
        task.wait_till_done()

    return _upload_invalid_data_rows_for_dataset


@pytest.fixture
def chat_evaluation_ontology(client, rand_gen):
    ontology_name = f"test-chat-evaluation-ontology-{rand_gen(str)}"
    ontology_builder = OntologyBuilder(
        tools=[
            Tool(tool=Tool.Type.MESSAGE_SINGLE_SELECTION,
                 name="model output single selection"),
            Tool(tool=Tool.Type.MESSAGE_MULTI_SELECTION,
                 name="model output multi selection"),
            Tool(tool=Tool.Type.MESSAGE_RANKING,
                 name="model output multi ranking"),
        ],
        classifications=[
            Classification(class_type=Classification.Type.TEXT,
                           name="global model output text classification",
                           scope=Classification.Scope.GLOBAL),
            Classification(class_type=Classification.Type.RADIO,
                           name="global model output radio classification",
                           scope=Classification.Scope.GLOBAL,
                           options=[
                               Option(value="global first option answer"),
                               Option(value="global second option answer"),
                           ]),
            Classification(class_type=Classification.Type.CHECKLIST,
                           name="global model output checklist classification",
                           scope=Classification.Scope.GLOBAL,
                           options=[
                               Option(value="global first option answer"),
                               Option(value="global second option answer"),
                           ]),
            Classification(class_type=Classification.Type.TEXT,
                           name="index model output text classification",
                           scope=Classification.Scope.INDEX),
            Classification(class_type=Classification.Type.RADIO,
                           name="index model output radio classification",
                           scope=Classification.Scope.INDEX,
                           options=[
                               Option(value="index first option answer"),
                               Option(value="index second option answer"),
                           ]),
            Classification(class_type=Classification.Type.CHECKLIST,
                           name="index model output checklist classification",
                           scope=Classification.Scope.INDEX,
                           options=[
                               Option(value="index first option answer"),
                               Option(value="index second option answer"),
                           ]),
        ])

    ontology = client.create_ontology(
        ontology_name,
        ontology_builder.asdict(),
        media_type=MediaType.Conversational,
        ontology_kind=OntologyKind.ModelEvaluation)

    yield ontology

    try:
        client.delete_unused_ontology(ontology.uid)
    except Exception as e:
        print(f"Failed to delete ontology {ontology.uid}: {str(e)}")


@pytest.fixture
def live_chat_evaluation_project_with_new_dataset(client, rand_gen):
    project_name = f"test-model-evaluation-project-{rand_gen(str)}"
    dataset_name = f"test-model-evaluation-dataset-{rand_gen(str)}"
    project = client.create_model_evaluation_project(name=project_name,
                                                     dataset_name=dataset_name,
                                                     data_row_count=1)

    yield project

    project.delete()


@pytest.fixture
def offline_chat_evaluation_project(client, rand_gen):
    project_name = f"test-offline-model-evaluation-project-{rand_gen(str)}"
    project = client.create_offline_model_evaluation_project(name=project_name)

    yield project

    project.delete()


@pytest.fixture
def chat_evaluation_project_append_to_dataset(client, dataset, rand_gen):
    project_name = f"test-model-evaluation-project-{rand_gen(str)}"
    dataset_id = dataset.uid
    project = client.create_model_evaluation_project(name=project_name,
                                                     dataset_id=dataset_id,
                                                     data_row_count=1)

    yield project

    project.delete()


@pytest.fixture
def offline_conversational_data_row(initial_dataset):
    convo_v2_row_data = {
        "type": "application/vnd.labelbox.conversational.model-chat-evaluation",
        "version": 2,
        "actors": {
            "clxhs9wk000013b6w7imiz0h8": {
                "role": "human",
                "metadata": {
                    "name": "User"
                }
            },
            "clxhsc6xb00013b6w1awh579j": {
                "role": "model",
                "metadata": {
                    "modelConfigId": "5a50d319-56bd-405d-87bb-4442daea0d0f"
                }
            },
            "clxhsc6xb00023b6wlp0768zs": {
                "role": "model",
                "metadata": {
                    "modelConfigId": "1cfc833a-2684-47df-95ac-bb7d9f9e3e1f"
                }
            }
        },
        "messages": {
            "clxhs9wk000023b6wrufora3k": {
                "actorId": "clxhs9wk000013b6w7imiz0h8",
                "content": [{
                    "type": "text",
                    "content": "Hello world"
                }],
                "childMessageIds": ["clxhscb4z00033b6wukpvmuol"]
            },
            "clxhscb4z00033b6wukpvmuol": {
                "actorId": "clxhsc6xb00013b6w1awh579j",
                "content": [{
                    "type":
                        "text",
                    "content":
                        "Hello to you too! 👋 \n\nIt's great to be your guide in the digital world. What can I help you with today? 😊 \n"
                }],
                "childMessageIds": ["clxhu2s0900013b6wbv0ndddd"]
            },
            "clxhu2s0900013b6wbv0ndddd": {
                "actorId":
                    "clxhs9wk000013b6w7imiz0h8",
                "content": [{
                    "type": "text",
                    "content": "Lets some some multi-turn happening"
                }],
                "childMessageIds": [
                    "clxhu4qib00023b6wuep47b1l", "clxhu4qib00033b6wf18az01q"
                ]
            },
            "clxhu4qib00023b6wuep47b1l": {
                "actorId": "clxhsc6xb00013b6w1awh579j",
                "content": [{
                    "type":
                        "text",
                    "content":
                        "Okay, I'm ready for some multi-turn fun! To make it interesting, how about we try building a story together? \n\n**Here's the beginning:**\n\nThe old, dusty book lay forgotten on the shelf, its leather cover cracked and faded. But as the afternoon sun slanted through the window, a single ray caught a glint of gold on the book's spine. Suddenly...\n\n**Now you tell me what happens!** What does the glint of gold turn out to be? What happens next? 🤔 \n"
                }],
                "childMessageIds": ["clxhu596m00043b6wvkgahcwz"]
            },
            "clxhu4qib00033b6wf18az01q": {
                "actorId": "clxhsc6xb00023b6wlp0768zs",
                "content": [{
                    "type":
                        "text",
                    "content":
                        "Okay, I'm game! Let's have a multi-turn conversation! But to get us started, give me a topic or a question to get us going! 😄 \n\nWhat's on your mind? \n"
                }],
                "childMessageIds": ["clxhu596m00043b6wvkgahcwz"]
            },
            "clxhu596m00043b6wvkgahcwz": {
                "actorId":
                    "clxhs9wk000013b6w7imiz0h8",
                "content": [{
                    "type": "text",
                    "content": "Can you generate an image?"
                }],
                "childMessageIds": [
                    "clxhu5hqk00053b6wk21u6xi7", "clxhu5hqk00063b6whl4gofx8"
                ]
            },
            "clxhu5hqk00053b6wk21u6xi7": {
                "actorId": "clxhsc6xb00013b6w1awh579j",
                "content": [{
                    "type":
                        "text",
                    "content":
                        "I can't generate images directly yet. I'm a text-based AI, so my superpowers are all about words! 🪄 \n\nHowever, I'd love to help you describe an image that fits our story. Tell me: \n\n* **What should the image show?** (For example: The dusty book, the glint of gold, the bookshelf, the room, etc.) \n* **What kind of feeling or atmosphere should the image have?** (Mysterious? Magical? Ominous?) \n\nThe more details you give me, the better I can describe the image! ✨ \n"
                }],
                "childMessageIds": []
            },
            "clxhu5hqk00063b6whl4gofx8": {
                "actorId": "clxhsc6xb00023b6wlp0768zs",
                "content": [{
                    "type":
                        "text",
                    "content":
                        "I can't *actually* generate images directly. 😔 I'm primarily a text-based AI. \n\nTo help me understand what you'd like to see, tell me: \n\n* **What should be in the image?** Be specific! (e.g., \"A cat wearing a tiny hat\", \"A futuristic cityscape at sunset\")\n* **What style do you imagine?** (e.g., realistic, cartoonish, abstract)\n\nOnce you give me those details, I can try to give you a vivid description that's almost as good as seeing it! 😊 \n"
                }],
                "childMessageIds": []
            }
        },
        "rootMessageIds": ["clxhs9wk000023b6wrufora3k"]
    }

    convo_v2_asset = {
        "row_data": convo_v2_row_data,
    }
    data_row = initial_dataset.create_data_row(convo_v2_asset)

    return data_row


@pytest.fixture()
def conversation_data_row(initial_dataset, rand_gen):
    data = {
        "row_data":
            "https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json",
        "global_key":
            f"https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json-{rand_gen(str)}",
    }
    convo_asset = {"row_data": data}
    data_row = initial_dataset.create_data_row(convo_asset)

    return data_row


def pytest_configure():
    pytest.report = defaultdict(int)


@pytest.hookimpl(hookwrapper=True)
def pytest_fixture_setup(fixturedef):
    start = time.time()
    yield
    end = time.time()

    exec_time = end - start
    if "FIXTURE_PROFILE" in os.environ:
        pytest.report[fixturedef.argname] += exec_time


@pytest.fixture(scope='session', autouse=True)
def print_perf_summary():
    yield

    if "FIXTURE_PROFILE" in os.environ:
        sorted_dict = dict(
            sorted(pytest.report.items(),
                   key=lambda item: item[1],
                   reverse=True))
        num_of_entries = 10 if len(sorted_dict) >= 10 else len(sorted_dict)
        slowest_fixtures = [(aaa, sorted_dict[aaa])
                            for aaa in islice(sorted_dict, num_of_entries)]
        print("\nTop slowest fixtures:\n", slowest_fixtures, file=sys.stderr)



import pytest

from labelbox.data.annotation_types.metrics import ConfusionMatrixAggregation, ScalarMetricAggregation
from labelbox.data.annotation_types.metrics import ConfusionMatrixMetric, ScalarMetric
from labelbox.data.annotation_types import ScalarMetric, Label, ImageData
from labelbox.data.annotation_types.metrics.scalar import RESERVED_METRIC_NAMES
from labelbox import pydantic_compat


def test_legacy_scalar_metric():
    value = 10
    metric = ScalarMetric(value=value)
    assert metric.value == value

    label = Label(data=ImageData(uid="ckrmd9q8g000009mg6vej7hzg"),
                  annotations=[metric])
    expected = {
        'data': {
            'external_id': None,
            'uid': 'ckrmd9q8g000009mg6vej7hzg',
            'global_key': None,
            'im_bytes': None,
            'file_path': None,
            'url': None,
            'arr': None,
            'media_attributes': None,
            'metadata': None,
        },
        'annotations': [{
            'value': 10.0,
            'extra': {},
        }],
        'extra': {},
        'uid': None
    }
    assert label.dict() == expected


# TODO: Test with confidence


@pytest.mark.parametrize('feature_name,subclass_name,aggregation,value', [
    ("cat", "orange", ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
    ("cat", None, ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
    (None, None, ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
    (None, None, None, 0.5),
    ("cat", "orange", ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
    ("cat", None, ScalarMetricAggregation.HARMONIC_MEAN, 0.5),
    (None, None, ScalarMetricAggregation.GEOMETRIC_MEAN, 0.5),
    (None, None, ScalarMetricAggregation.SUM, 0.5),
    ("cat", "orange", ScalarMetricAggregation.ARITHMETIC_MEAN, {
        0.1: 0.2,
        0.3: 0.5,
        0.4: 0.8
    }),
])
def test_custom_scalar_metric(feature_name, subclass_name, aggregation, value):
    kwargs = {'aggregation': aggregation} if aggregation is not None else {}
    metric = ScalarMetric(metric_name="custom_iou",
                          value=value,
                          feature_name=feature_name,
                          subclass_name=subclass_name,
                          **kwargs)
    assert metric.value == value

    label = Label(data=ImageData(uid="ckrmd9q8g000009mg6vej7hzg"),
                  annotations=[metric])
    expected = {
        'data': {
            'external_id': None,
            'uid': 'ckrmd9q8g000009mg6vej7hzg',
            'global_key': None,
            'im_bytes': None,
            'file_path': None,
            'url': None,
            'arr': None,
            'media_attributes': None,
            'metadata': None,
        },
        'annotations': [{
            'value':
                value,
            'metric_name':
                'custom_iou',
            **({
                'feature_name': feature_name
            } if feature_name else {}),
            **({
                'subclass_name': subclass_name
            } if subclass_name else {}), 'aggregation':
                aggregation or ScalarMetricAggregation.ARITHMETIC_MEAN,
            'extra': {}
        }],
        'extra': {},
        'uid': None
    }

    assert label.dict() == expected


@pytest.mark.parametrize('feature_name,subclass_name,aggregation,value', [
    ("cat", "orange", ConfusionMatrixAggregation.CONFUSION_MATRIX,
     (0, 1, 2, 3)),
    ("cat", None, ConfusionMatrixAggregation.CONFUSION_MATRIX, (0, 1, 2, 3)),
    (None, None, ConfusionMatrixAggregation.CONFUSION_MATRIX, (0, 1, 2, 3)),
    (None, None, None, (0, 1, 2, 3)),
    ("cat", "orange", ConfusionMatrixAggregation.CONFUSION_MATRIX, {
        0.1: (0, 1, 2, 3),
        0.3: (0, 1, 2, 3),
        0.4: (0, 1, 2, 3)
    }),
])
def test_custom_confusison_matrix_metric(feature_name, subclass_name,
                                         aggregation, value):
    kwargs = {'aggregation': aggregation} if aggregation is not None else {}
    metric = ConfusionMatrixMetric(metric_name="confusion_matrix_50_pct_iou",
                                   value=value,
                                   feature_name=feature_name,
                                   subclass_name=subclass_name,
                                   **kwargs)
    assert metric.value == value

    label = Label(data=ImageData(uid="ckrmd9q8g000009mg6vej7hzg"),
                  annotations=[metric])
    expected = {
        'data': {
            'external_id': None,
            'uid': 'ckrmd9q8g000009mg6vej7hzg',
            'global_key': None,
            'im_bytes': None,
            'file_path': None,
            'url': None,
            'arr': None,
            'media_attributes': None,
            'metadata': None,
        },
        'annotations': [{
            'value':
                value,
            'metric_name':
                'confusion_matrix_50_pct_iou',
            **({
                'feature_name': feature_name
            } if feature_name else {}),
            **({
                'subclass_name': subclass_name
            } if subclass_name else {}), 'aggregation':
                aggregation or ConfusionMatrixAggregation.CONFUSION_MATRIX,
            'extra': {}
        }],
        'extra': {},
        'uid': None
    }
    assert label.dict() == expected


def test_name_exists():
    # Name is only required for ConfusionMatrixMetric for now.
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(value=[0, 1, 2, 3])
    assert "field required (type=value_error.missing)" in str(exc_info.value)


def test_invalid_aggregations():
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ScalarMetric(
            metric_name="invalid aggregation",
            value=0.1,
            aggregation=ConfusionMatrixAggregation.CONFUSION_MATRIX)
    assert "value is not a valid enumeration member" in str(exc_info.value)
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(metric_name="invalid aggregation",
                                       value=[0, 1, 2, 3],
                                       aggregation=ScalarMetricAggregation.SUM)
    assert "value is not a valid enumeration member" in str(exc_info.value)


def test_invalid_number_of_confidence_scores():
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ScalarMetric(metric_name="too few scores", value={0.1: 0.1})
    assert "Number of confidence scores must be greater" in str(exc_info.value)
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(metric_name="too few scores",
                                       value={0.1: [0, 1, 2, 3]})
    assert "Number of confidence scores must be greater" in str(exc_info.value)
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ScalarMetric(metric_name="too many scores",
                              value={i / 20.: 0.1 for i in range(20)})
    assert "Number of confidence scores must be greater" in str(exc_info.value)
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(
            metric_name="too many scores",
            value={i / 20.: [0, 1, 2, 3] for i in range(20)})
    assert "Number of confidence scores must be greater" in str(exc_info.value)


@pytest.mark.parametrize("metric_name", RESERVED_METRIC_NAMES)
def test_reserved_names(metric_name: str):
    with pytest.raises(pydantic_compat.ValidationError) as exc_info:
        ScalarMetric(metric_name=metric_name, value=0.5)
    assert 'is a reserved metric name' in exc_info.value.errors()[0]['msg']



import pytest

from labelbox.exceptions import InconsistentOntologyException
from labelbox import Tool, Classification, Option, OntologyBuilder
from itertools import product

_SAMPLE_ONTOLOGY = {
    "tools": [{
        "schemaNodeId": None,
        "featureSchemaId": None,
        "required": False,
        "name": "poly",
        "color": "#FF0000",
        "tool": "polygon",
        "classifications": []
    }, {
        "schemaNodeId": None,
        "featureSchemaId": None,
        "required": False,
        "name": "segment",
        "color": "#FF0000",
        "tool": "superpixel",
        "classifications": []
    }, {
        "schemaNodeId":
            None,
        "featureSchemaId":
            None,
        "required":
            False,
        "name":
            "bbox",
        "color":
            "#FF0000",
        "tool":
            "rectangle",
        "classifications": [{
            "schemaNodeId":
                None,
            "featureSchemaId":
                None,
            "required":
                True,
            "instructions":
                "nested classification",
            "name":
                "nested classification",
            "type":
                "radio",
            'uiMode': 
                "searchable",
            "options": [{
                "schemaNodeId":
                    None,
                "featureSchemaId":
                    None,
                "label":
                    "first",
                "value":
                    "first",
                "options": [{
                    "schemaNodeId": None,
                    "featureSchemaId": None,
                    "required": False,
                    "instructions": "nested nested text",
                    "name": "nested nested text",
                    "type": "text",
                    "options": []
                }]
            }, {
                "schemaNodeId": None,
                "featureSchemaId": None,
                "label": "second",
                "value": "second",
                "options": []
            }]
        }, {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": True,
            "instructions": "nested text",
            "name": "nested text",
            "type": "text",
            "options": []
        }]
    }, {
        "schemaNodeId": None,
        "featureSchemaId": None,
        "required": False,
        "name": "dot",
        "color": "#FF0000",
        "tool": "point",
        "classifications": []
    }, {
        "schemaNodeId": None,
        "featureSchemaId": None,
        "required": False,
        "name": "polyline",
        "color": "#FF0000",
        "tool": "line",
        "classifications": []
    }, {
        "schemaNodeId": None,
        "featureSchemaId": None,
        "required": False,
        "name": "ner",
        "color": "#FF0000",
        "tool": "named-entity",
        "classifications": []
    }],
    "classifications": [{
        "schemaNodeId":
            None,
        "featureSchemaId":
            None,
        "required":
            True,
        "instructions":
            "This is a question.",
        "name":
            "This is a question.",
        "type":
            "radio",
        "scope":
            "global",
        'uiMode': 
            "searchable",
        "options": [{
            "schemaNodeId": None,
            "featureSchemaId": None,
            "label": "yes",
            "value": "definitely yes",
            "options": []
        }, {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "label": "no",
            "value": "definitely not",
            "options": []
        }]
    }]
}


@pytest.mark.parametrize("tool_type", list(Tool.Type))
def test_create_tool(tool_type) -> None:
    t = Tool(tool=tool_type, name="tool")
    assert (t.tool == tool_type)


@pytest.mark.parametrize("class_type", list(Classification.Type))
def test_create_classification(class_type) -> None:
    c = Classification(class_type=class_type, name="classification")
    assert (c.class_type == class_type)

@pytest.mark.parametrize("ui_mode_type, class_type", list(product(list(Classification.UIMode), list(Classification.Type))))
def test_create_classification_with_ui_mode(ui_mode_type, class_type) -> None:
    c = Classification(name="classification", class_type=class_type, ui_mode=ui_mode_type)
    assert (c.ui_mode == ui_mode_type)


@pytest.mark.parametrize("value, expected_value, typing",
                         [(3, 3, int), ("string", "string", str)])
def test_create_option_with_value(value, expected_value, typing) -> None:
    o = Option(value=value)
    assert (o.value == expected_value)
    assert (o.value == o.label)


@pytest.mark.parametrize("value, label, expected_value, typing",
                         [(3, 2, 3, int),
                          ("string", "another string", "string", str)])
def test_create_option_with_value_and_label(value, label, expected_value,
                                            typing) -> None:
    o = Option(value=value, label=label)
    assert (o.value == expected_value)
    assert o.value != o.label
    assert isinstance(o.value, typing)


def test_create_empty_ontology() -> None:
    o = OntologyBuilder()
    assert (o.tools == [])
    assert (o.classifications == [])


def test_add_ontology_tool() -> None:
    o = OntologyBuilder()
    o.add_tool(Tool(tool=Tool.Type.BBOX, name="bounding box"))

    second_tool = Tool(tool=Tool.Type.SEGMENTATION, name="segmentation")
    o.add_tool(second_tool)
    assert len(o.tools) == 2

    for tool in o.tools:
        assert (type(tool) == Tool)

    with pytest.raises(InconsistentOntologyException) as exc:
        o.add_tool(Tool(tool=Tool.Type.BBOX, name="bounding box"))
    assert "Duplicate tool name" in str(exc.value)


def test_add_ontology_classification() -> None:
    o = OntologyBuilder()
    o.add_classification(
        Classification(class_type=Classification.Type.TEXT, name="text"))

    second_classification = Classification(
        class_type=Classification.Type.CHECKLIST, name="checklist")
    o.add_classification(second_classification)
    assert len(o.classifications) == 2

    for classification in o.classifications:
        assert (type(classification) == Classification)

    with pytest.raises(InconsistentOntologyException) as exc:
        o.add_classification(
            Classification(class_type=Classification.Type.TEXT, name="text"))
    assert "Duplicate classification name" in str(exc.value)


def test_tool_add_classification() -> None:
    t = Tool(tool=Tool.Type.SEGMENTATION, name="segmentation")
    c = Classification(class_type=Classification.Type.TEXT, name="text")
    t.add_classification(c)
    assert t.classifications == [c]

    with pytest.raises(Exception) as exc:
        t.add_classification(c)
    assert "Duplicate nested classification" in str(exc)


def test_classification_add_option() -> None:
    c = Classification(class_type=Classification.Type.RADIO, name="radio")
    o = Option(value="option")
    c.add_option(o)
    assert c.options == [o]

    with pytest.raises(InconsistentOntologyException) as exc:
        c.add_option(Option(value="option"))
    assert "Duplicate option" in str(exc.value)


def test_option_add_option() -> None:
    o = Option(value="option")
    c = Classification(class_type=Classification.Type.TEXT, name="text")
    o.add_option(c)
    assert o.options == [c]

    with pytest.raises(InconsistentOntologyException) as exc:
        o.add_option(c)
    assert "Duplicate nested classification" in str(exc.value)


def test_ontology_asdict() -> None:
    assert OntologyBuilder.from_dict(
        _SAMPLE_ONTOLOGY).asdict() == _SAMPLE_ONTOLOGY


def test_classification_using_instructions_instead_of_name_shows_warning():
    with pytest.warns(Warning):
        Classification(class_type=Classification.Type.TEXT, instructions="text")


def test_classification_without_name_raises_error():
    with pytest.raises(ValueError):
        Classification(class_type=Classification.Type.TEXT)



import pytest

from labelbox.exceptions import LabelboxError, OperationNotAllowedException


def test_live_chat_evaluation_project(
        live_chat_evaluation_project_with_new_dataset, model_config):

    project = live_chat_evaluation_project_with_new_dataset

    project.set_project_model_setup_complete()
    assert bool(project.model_setup_complete) is True

    with pytest.raises(
            expected_exception=LabelboxError,
            match=
            "Cannot create model config for project because model setup is complete"
    ):
        project.add_model_config(model_config.uid)


def test_live_chat_evaluation_project_delete_cofig(
        live_chat_evaluation_project_with_new_dataset, model_config):

    project = live_chat_evaluation_project_with_new_dataset
    project_model_config_id = project.add_model_config(model_config.uid)
    assert project_model_config_id

    project_model_config = None
    for pmg in project.project_model_configs():
        if pmg.uid == project_model_config_id:
            project_model_config = pmg
            break
    assert project_model_config

    project.set_project_model_setup_complete()
    assert bool(project.model_setup_complete) is True

    with pytest.raises(
            expected_exception=LabelboxError,
            match=
            "Cannot create model config for project because model setup is complete"
    ):
        project_model_config.delete()


def test_offline_chat_evaluation_project(offline_chat_evaluation_project,
                                         model_config):

    project = offline_chat_evaluation_project

    with pytest.raises(
            expected_exception=OperationNotAllowedException,
            match=
            "Only live model chat evaluation projects can complete model setup"
    ):
        project.set_project_model_setup_complete()


def test_any_other_project(project, model_config):
    with pytest.raises(
            expected_exception=OperationNotAllowedException,
            match=
            "Only live model chat evaluation projects can complete model setup"
    ):
        project.set_project_model_setup_complete()



import uuid
import time
import pytest
from labelbox.schema.queue_mode import QueueMode
from labelbox.schema.media_type import MediaType
from labelbox.schema.labeling_frontend import LabelingFrontend
from labelbox.schema.annotation_import import LabelImport, AnnotationImportState


@pytest.fixture
def ontology():
    bbox_tool_with_nested_text = {
        'required':
            False,
        'name':
            'bbox_tool_with_nested_text',
        'tool':
            'rectangle',
        'color':
            '#a23030',
        'classifications': [{
            'required':
                False,
            'instructions':
                'nested',
            'name':
                'nested',
            'type':
                'radio',
            'options': [{
                'label':
                    'radio_option_1',
                'value':
                    'radio_value_1',
                'options': [{
                    'required':
                        False,
                    'instructions':
                        'nested_checkbox',
                    'name':
                        'nested_checkbox',
                    'type':
                        'checklist',
                    'options': [{
                        'label': 'nested_checkbox_option_1',
                        'value': 'nested_checkbox_value_1',
                        'options': []
                    }, {
                        'label': 'nested_checkbox_option_2',
                        'value': 'nested_checkbox_value_2'
                    }]
                }, {
                    'required': False,
                    'instructions': 'nested_text',
                    'name': 'nested_text',
                    'type': 'text',
                    'options': []
                }]
            },]
        }]
    }

    bbox_tool = {
        'required':
            False,
        'name':
            'bbox',
        'tool':
            'rectangle',
        'color':
            '#a23030',
        'classifications': [{
            'required':
                False,
            'instructions':
                'nested',
            'name':
                'nested',
            'type':
                'radio',
            'options': [{
                'label':
                    'radio_option_1',
                'value':
                    'radio_value_1',
                'options': [{
                    'required':
                        False,
                    'instructions':
                        'nested_checkbox',
                    'name':
                        'nested_checkbox',
                    'type':
                        'checklist',
                    'options': [{
                        'label': 'nested_checkbox_option_1',
                        'value': 'nested_checkbox_value_1',
                        'options': []
                    }, {
                        'label': 'nested_checkbox_option_2',
                        'value': 'nested_checkbox_value_2'
                    }]
                }]
            },]
        }]
    }

    polygon_tool = {
        'required': False,
        'name': 'polygon',
        'tool': 'polygon',
        'color': '#FF34FF',
        'classifications': []
    }
    polyline_tool = {
        'required': False,
        'name': 'polyline',
        'tool': 'line',
        'color': '#FF4A46',
        'classifications': []
    }
    point_tool = {
        'required': False,
        'name': 'point--',
        'tool': 'point',
        'color': '#008941',
        'classifications': []
    }
    entity_tool = {
        'required': False,
        'name': 'entity--',
        'tool': 'named-entity',
        'color': '#006FA6',
        'classifications': []
    }
    segmentation_tool = {
        'required': False,
        'name': 'segmentation--',
        'tool': 'superpixel',
        'color': '#A30059',
        'classifications': []
    }
    raster_segmentation_tool = {
        'required': False,
        'name': 'segmentation_mask',
        'tool': 'raster-segmentation',
        'color': '#ff0000',
        'classifications': []
    }
    checklist = {
        'required':
            False,
        'instructions':
            'checklist',
        'name':
            'checklist',
        'type':
            'checklist',
        'options': [{
            'label': 'option1',
            'value': 'option1'
        }, {
            'label': 'option2',
            'value': 'option2'
        }, {
            'label': 'optionN',
            'value': 'optionn'
        }]
    }
    checklist_index = {
        'required':
            False,
        'instructions':
            'checklist_index',
        'name':
            'checklist_index',
        'type':
            'checklist',
        'scope':
            'index',
        'options': [{
            'label': 'option1_index',
            'value': 'option1_index'
        }, {
            'label': 'option2_index',
            'value': 'option2_index'
        }, {
            'label': 'optionN_index',
            'value': 'optionn_index'
        }]
    }
    free_form_text = {
        'required': False,
        'instructions': 'text',
        'name': 'text',
        'type': 'text',
        'options': []
    }
    free_form_text_index = {
        'required': False,
        'instructions': 'text_index',
        'name': 'text_index',
        'type': 'text',
        'scope': 'index',
        'options': []
    }
    radio = {
        'required':
            False,
        'instructions':
            'radio',
        'name':
            'radio',
        'type':
            'radio',
        'options': [{
            'label': 'first_radio_answer',
            'value': 'first_radio_answer',
            'options': []
        }, {
            'label': 'second_radio_answer',
            'value': 'second_radio_answer',
            'options': []
        }]
    }
    named_entity = {
        'tool': 'named-entity',
        'name': 'named-entity',
        'required': False,
        'color': '#A30059',
        'classifications': [],
    }

    tools = [
        bbox_tool,
        bbox_tool_with_nested_text,
        polygon_tool,
        polyline_tool,
        point_tool,
        entity_tool,
        segmentation_tool,
        raster_segmentation_tool,
        named_entity,
    ]
    classifications = [
        checklist, checklist_index, free_form_text, free_form_text_index, radio
    ]
    return {"tools": tools, "classifications": classifications}


@pytest.fixture
def polygon_inference(prediction_id_mapping):
    polygon = prediction_id_mapping['polygon'].copy()
    polygon.update({
        "polygon": [{
            "x": 147.692,
            "y": 118.154
        }, {
            "x": 142.769,
            "y": 104.923
        }, {
            "x": 57.846,
            "y": 118.769
        }, {
            "x": 28.308,
            "y": 169.846
        }]
    })
    del polygon['tool']
    return polygon


@pytest.fixture
def configured_project_with_ontology(client, initial_dataset, ontology,
                                     rand_gen, image_url):
    dataset = initial_dataset
    project = client.create_project(
        name=rand_gen(str),
        queue_mode=QueueMode.Batch,
    )
    editor = list(
        client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]
    project.setup(editor, ontology)
    data_row_ids = []

    for _ in range(len(ontology['tools']) + len(ontology['classifications'])):
        data_row_ids.append(dataset.create_data_row(row_data=image_url).uid)
    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids
    yield project
    project.delete()


@pytest.fixture
def configured_project_without_data_rows(client, ontology, rand_gen):
    project = client.create_project(name=rand_gen(str),
                                    description=rand_gen(str),
                                    queue_mode=QueueMode.Batch)
    editor = list(
        client.get_labeling_frontends(
            where=LabelingFrontend.name == "editor"))[0]
    project.setup(editor, ontology)
    yield project
    project.delete()


@pytest.fixture
def model_run_with_data_rows(client, configured_project_with_ontology,
                             model_run_predictions, model_run,
                             wait_for_label_processing):
    configured_project_with_ontology.enable_model_assisted_labeling()
    use_data_row_ids = [p['dataRow']['id'] for p in model_run_predictions]
    model_run.upsert_data_rows(use_data_row_ids)

    upload_task = LabelImport.create_from_objects(
        client, configured_project_with_ontology.uid,
        f"label-import-{uuid.uuid4()}", model_run_predictions)
    upload_task.wait_until_done()
    assert upload_task.state == AnnotationImportState.FINISHED, "Label Import did not finish"
    assert len(
        upload_task.errors
    ) == 0, f"Label Import {upload_task.name} failed with errors {upload_task.errors}"
    labels = wait_for_label_processing(configured_project_with_ontology)
    label_ids = [label.uid for label in labels]
    model_run.upsert_labels(label_ids)
    yield model_run, labels
    model_run.delete()
    # TODO: Delete resources when that is possible ..


@pytest.fixture
def model_run_predictions(polygon_inference, rectangle_inference,
                          line_inference):
    # Not supporting mask since there isn't a signed url representing a seg mask to upload
    return [polygon_inference, rectangle_inference, line_inference]


@pytest.fixture
def model(client, rand_gen, configured_project):
    ontology = configured_project.ontology()
    data = {"name": rand_gen(str), "ontology_id": ontology.uid}
    model = client.create_model(data["name"], data["ontology_id"])
    yield model
    try:
        model.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def model_run(rand_gen, model):
    name = rand_gen(str)
    model_run = model.create_model_run(name)
    yield model_run
    try:
        model_run.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def wait_for_label_processing():
    """
    Do not use. Only for testing.

    Returns project's labels as a list after waiting for them to finish processing.
    If `project.labels()` is called before label is fully processed,
    it may return an empty set
    """

    def func(project):
        timeout_seconds = 10
        while True:
            labels = list(project.labels())
            if len(labels) > 0:
                return labels
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for label for project '{project.uid}' to finish processing"
                )
            time.sleep(2)

    return func


@pytest.fixture
def prediction_id_mapping(configured_project_with_ontology):
    # Maps tool types to feature schema ids
    project = configured_project_with_ontology
    ontology = project.ontology().normalized
    result = {}

    for idx, tool in enumerate(ontology['tools'] + ontology['classifications']):
        if 'tool' in tool:
            tool_type = tool['tool']
        else:
            tool_type = tool[
                'type'] if 'scope' not in tool else f"{tool['type']}_{tool['scope']}"  # so 'checklist' of 'checklist_index'

        # TODO: remove this once we have a better way to associate multiple tools instances with a single tool type
        if tool_type == 'rectangle':
            value = {
                "uuid": str(uuid.uuid4()),
                "schemaId": tool['featureSchemaId'],
                "name": tool['name'],
                "dataRow": {
                    "id": project.data_row_ids[idx],
                },
                'tool': tool
            }
            if tool_type not in result:
                result[tool_type] = []
            result[tool_type].append(value)
        else:
            result[tool_type] = {
                "uuid": str(uuid.uuid4()),
                "schemaId": tool['featureSchemaId'],
                "name": tool['name'],
                "dataRow": {
                    "id": project.data_row_ids[idx],
                },
                'tool': tool
            }
    return result


@pytest.fixture
def line_inference(prediction_id_mapping):
    line = prediction_id_mapping['line'].copy()
    line.update(
        {"line": [{
            "x": 147.692,
            "y": 118.154
        }, {
            "x": 150.692,
            "y": 160.154
        }]})
    del line['tool']
    return line


@pytest.fixture
def polygon_inference(prediction_id_mapping):
    polygon = prediction_id_mapping['polygon'].copy()
    polygon.update({
        "polygon": [{
            "x": 147.692,
            "y": 118.154
        }, {
            "x": 142.769,
            "y": 104.923
        }, {
            "x": 57.846,
            "y": 118.769
        }, {
            "x": 28.308,
            "y": 169.846
        }]
    })
    del polygon['tool']
    return polygon


def find_tool_by_name(tool_instances, name):
    for tool in tool_instances:
        if tool['name'] == name:
            return tool
    return None


@pytest.fixture
def rectangle_inference(prediction_id_mapping):
    tool_instance = find_tool_by_name(prediction_id_mapping['rectangle'],
                                      'bbox')
    rectangle = tool_instance.copy()
    rectangle.update({
        "bbox": {
            "top": 48,
            "left": 58,
            "height": 65,
            "width": 12
        },
        'classifications': [{
            "schemaId":
                rectangle['tool']['classifications'][0]['featureSchemaId'],
            "name":
                rectangle['tool']['classifications'][0]['name'],
            "answer": {
                "schemaId":
                    rectangle['tool']['classifications'][0]['options'][0]
                    ['featureSchemaId'],
                "name":
                    rectangle['tool']['classifications'][0]['options'][0]
                    ['value']
            }
        }]
    })
    del rectangle['tool']
    return rectangle
